/***************************************************************
 *  bst.c -- a C++ program to construct and maintain a  binary *
 *           search tree annotated with SAIL function calls    *
 ***************************************************************/

/* 
   To get the executable file without SWAN annotation:

        g++ -o bst bst.c

   To get the executable file with SWAN annotation:

        g++ -o bst -DSWAN bst.c -lsail++ -lXt -lX11 -lm 

   assuming sail.h and libsail++.a are in current directory.
*/

#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <assert.h>

#define SWAN 1	//this is a test to see if I can get SWAN components

#ifdef SWAN
#include "sail.h"
#endif

#define FALSE 0
#define TRUE 1
#define DEFAULT_SIZE 10  // size for lists if no size is given.

template <class T> class BinNode {
  public:
    T element;          // The node's value
    BinNode<T>* left;   // Pointer to left child
    BinNode<T>* right;  // Pointer to right child

    BinNode() { left = right = NULL; }
    BinNode(T& e, BinNode<T>* l =NULL, BinNode<T>* r =NULL)
        { element = e; left = l; right = r; }
    ~BinNode() { }
    BinNode<T>* leftchild() { return left; }
    BinNode<T>* rightchild() { return right; }
    T& value() { return element; }
    BinNode<T>* setValue(T& val) { element = val; return this;}
    bool isLeaf() { return (left == NULL) && (right == NULL); }
};

template <class T> class BST {
  private:
    BinNode<T>* root;
    void clearhelp(BinNode<T>*);
    void inserthelp(BinNode<T>*&, T&);
    void removehelp(BinNode<T>*&, T&);
    BinNode<T>* findhelp(BinNode<T>*, T&);
    void printhelp(BinNode<T>*, int) const;
  public:
    BST() { root = NULL; }
    ~BST() { clearhelp(root); }
    BST<T>& clear()
        { clearhelp(root); root = NULL; return *this; }
    BST<T>& insert(T val)
        { inserthelp(root, val); return *this; }
    BST<T>& remove(T val)
        { removehelp(root, val); return *this; }
    BinNode<T>* deletemin(BinNode<T>*&);
    BinNode<T>* find(T val)
        { return findhelp(root, val); }
    bool isEmpty() const			///4 Feb 1997  changed bool to boolean 
        { return root == NULL; }
    void print() const {
        if (root == NULL) cout << "The BST is empty.\n";
        else printhelp(root, 0);
    }
    BinNode<T>* getRoot(void) { return root ; }
} ;

template <class T>
void BST<T>::clearhelp(BinNode<T>* rt) {
    if (rt == NULL) return;
    clearhelp(rt->leftchild());
    clearhelp(rt->rightchild());
    delete rt;
}

template <class T>
void BST<T>::removehelp(BinNode<T>*& rt, T& val) {
    if (rt == NULL) cout << val << " is not in the tree.\n";
    else if (val < rt->value()) removehelp(rt->left, val);
    else if (val > rt->value()) removehelp(rt->right, val);
    else {  // Found it -- now delete it
        BinNode<T>* temp = rt;
        if (rt->left == NULL) rt = rt->right;
        else if (rt->right == NULL) rt = rt->left;
        else {  // Both children are non-empty
            temp = deletemin(rt->right);
      	    rt->setValue(temp->value());
        }
        delete temp;
    }
}

template <class T>
BinNode<T>* BST<T>::deletemin(BinNode<T>*& rt) {
    BinNode<T>* temp;
    assert(rt != NULL);
    if (rt->left != NULL) return deletemin(rt->left);
    else { temp = rt; rt = rt->right; return temp; }
}

template <class T>
BinNode<T>* BST<T>::findhelp(BinNode<T>* rt, T& val) {
    if (rt == NULL) return NULL;
    else if (val < rt->value())
        return findhelp(rt->leftchild(), val);
    else if (val == rt->value()) return rt;
    else return findhelp(rt->rightchild(), val);
}

template <class T>
void BST<T>::inserthelp(BinNode<T>*& rt, T& val) {
  if (rt == NULL) rt = new BinNode<T>(val, NULL, NULL);
  else if (val < rt->value()) inserthelp(rt->left, val);
  else inserthelp(rt->right, val);
}

template <class T>
void BST<T>::printhelp(BinNode<T> *rt, int level) const {
    int i;
    if (rt == NULL) return;
    printhelp(rt->leftchild(), level+1);
    for (i=0; i<level; i++) cout << "  ";  // indent
        cout << rt->value() << "\n";
    printhelp(rt->rightchild(), level+1);
}

/***************************************************************
 * End of bst.c                                                *
 ***************************************************************/
